#! /usr/bin/env python
import sys
import os
import netaddr
import argparse
import ConfigParser
import requests
from subprocess import call, check_output, CalledProcessError

sys.path.append('/usr/lib/python2.7/dist-packages/contrail_vrouter_api/gen_py')
sys.path.append('/usr/share/contrail-utils')

from vnc_api import vnc_api
import cfgm_common.exceptions

import instance_service.ttypes
from contrail_utils import vrouter_rpc, uuid_from_string, ProcessExecutionError, format_dict
from contrail_vrouter_api.vrouter_api import ContrailVRouterApi


class MockPort(object):
    def __init__(self, *argv, **args):
        self.args = self._parse_args()
        self.vnc_lib = None
        try:
            self.vnc_lib = vnc_api.VncApi(username=self.args['auth_username'],
                                          password=self.args['auth_password'],
                                          tenant_name=self.args['auth_tenant'],
                                          api_server_host=self.args['api_host'],
                                          api_server_port=self.args['api_port'],
                                          auth_protocol=self.args['auth_protocol'],
                                          auth_host=self.args['auth_host'],
                                          auth_port=self.args['auth_port'],
                                          auth_type="keystone")
        except requests.exceptions.ConnectionError:
            print('VNC API connection failed. API Server={}:{}'.format(
                self.args['api_host'], self.args['api_port']))
            sys.exit(1)
        except RuntimeError as e:
            print(e.message)
            sys.exit(1)
        self.vrouter_api = ContrailVRouterApi()
        self.hostname = check_output(['hostname']).strip()

        self.proj_fq_name = ['default-domain', self.args['project_name']]
        self.proj_obj = self.get_proj_object(self.proj_fq_name)

        self.ipams = self.vnc_lib.network_ipams_list(parent_fq_name=self.proj_fq_name)
        self.ipam_fq_names = [x['fq_name'] for x in self.ipams['network-ipams']]
        self.vns = self.vnc_lib.virtual_networks_list(parent_fq_name=self.proj_fq_name)
        self.vn_fq_names = [x['fq_name'] for x in self.vns['virtual-networks']]
        self.vms = self.vnc_lib.virtual_machines_list()
        self.vm_fq_names = [x['fq_name'] for x in self.vms['virtual-machines']]
        self.vmis = self.vnc_lib.virtual_machine_interfaces_list(parent_fq_name=self.proj_fq_name)
        self.vmi_fq_names = [x['fq_name'] for x in self.vmis['virtual-machine-interfaces']]
        self.inst_ips = self.vnc_lib.instance_ips_list()
        self.inst_ip_fq_names = [x['fq_name'] for x in self.inst_ips['instance-ips']]

        tap_list_cmd = "ip link | grep ': tap' | awk {'print $2'} | sed -e 's/:$//g'"
        try:
            self.tap_list = check_output(tap_list_cmd, shell=True).splitlines()
        except CalledProcessError:
            self.tap_list = []
        
        netns_list_cmd = "ip netns list | grep ^{}_".format(self.args['project_name'])
        try:
            self.netns_list = check_output(netns_list_cmd, shell=True).splitlines()
        except CalledProcessError:
            self.netns_list = []
        
    def _parse_args(self):
        conf_parser = argparse.ArgumentParser(add_help=False)
        conf_parser.add_argument('--conf_file', default='mock_port.conf',
                                   help='INI-style config file', metavar='FILE')
        args, remaining_argv = conf_parser.parse_known_args()
        defaults = {
            "api_host": "127.0.0.1",
            "api_port": 8082,
            "auth_protocol": "http",
            "auth_host": "127.0.0.1",
            "auth_port": 5000,
            "auth_tenant": "admin",
            "auth_username": "admin",
            "auth_password": "password",
            "project_name": "demo",
            "num_vn": 1,
            "num_port_per_vn": 2,
            'config_only': False,
            'clear': False,
        }
        if os.path.isfile(args.conf_file):
            config = ConfigParser.SafeConfigParser()
            config.read([args.conf_file])
            for key, value in config.items('DEFAULT'):
                if key in ['api_port', 'auth_port', 'num_vn', 'num_port_per_vn']:
                    defaults[key] = int(value)
                else:
                    defaults[key] = value

        parser = argparse.ArgumentParser(
            parents=[conf_parser],
            description=__doc__,
            formatter_class=argparse.RawDescriptionHelpFormatter
        )
        parser.set_defaults(**defaults)
        parser.add_argument('--api_host', metavar='ADDRESS',
                            help='Config API server address')
        parser.add_argument('--api_port', type=int, metavar='PORT',
                            help='Config API server port')
        parser.add_argument('--auth_protocol', metavar='http|https',
                            help='Keystone auth protocol')
        parser.add_argument('--auth_host', metavar='ADDRESS',
                            help='Keystone auth address')
        parser.add_argument('--auth_port', type=int, metavar='PORT',
                            help='Keystone auth port')
        parser.add_argument('--auth_tenant', metavar='TENANT',
                            help='Keystone auth tenant name')
        parser.add_argument('--auth_username', metavar='USERNAME',
                            help='Keystone auth username')
        parser.add_argument('--auth_password', metavar='PASSWORD',
                            help='Keystone auth password')
        parser.add_argument('--project_name', metavar='PROJECT',
                            help='Project name where to create veth port')
        parser.add_argument('--num_vn', type=int, metavar='NUM',
                            help='Number of virtual-networks')
        parser.add_argument('--num_port_per_vn', type=int, metavar='NUM',
                            help='Number of port per virtual-network')
        parser.add_argument('--config_only', action='store_true',
                            help='Do not create veth port')
        parser.add_argument('--clear', action='store_true',
                            help='Clean up mock port objects')
        args = vars(parser.parse_args(remaining_argv))
        if len(args['project_name']) > 8:
            print('Project name must be less than 8 characters.')
            sys.exit(1)
        if args['num_vn'] > 10000:
            print('Number of virtual-network must be less than 10,000.')
            sys.exit(1)
        if args['num_port_per_vn'] > 100:
            print('Number of port per virtual-network must be less than 100.')
            sys.exit(1)
        return args

    def get_proj_object(self, fq_name):
        fq_names = [x['fq_name'] for x in self.vnc_lib.projects_list()['projects']]
        if fq_name not in fq_names:
            pj = vnc_api.Project(name=fq_name[1], fq_name=fq_name)
            print('Creating {}'.format(fq_name))
            self.vnc_lib.project_create(pj)
        proj = self.vnc_lib.project_read(fq_name=fq_name)
        return proj

    def create_vn(self, vn_name, subnet_addr, subnet_len):
        ipam_name = 'ipam_' + vn_name
        ipam_fq_name = self.proj_fq_name + [ipam_name]
        if ipam_fq_name not in self.ipam_fq_names:
            ipam = vnc_api.NetworkIpam(name=ipam_name, fq_name=ipam_fq_name, parent_obj=self.proj_obj)
            self.vnc_lib.network_ipam_create(ipam)
        ipam = self.vnc_lib.network_ipam_read(fq_name=ipam_fq_name)

        vn_fq_name = self.proj_fq_name + [vn_name]
        if vn_fq_name not in self.vn_fq_names:
            subnet = vnc_api.SubnetType(ip_prefix=subnet_addr, ip_prefix_len=int(subnet_len))
            ipam_subnet = vnc_api.IpamSubnetType(subnet=subnet)
            vn = vnc_api.VirtualNetwork(name=vn_name, fq_name=vn_fq_name, parent_obj=self.proj_obj)
            vn.set_network_ipam(ref_obj=ipam, ref_data=vnc_api.VnSubnetsType([ipam_subnet]))
            self.vnc_lib.virtual_network_create(vn)
        self.vn = self.vnc_lib.virtual_network_read(fq_name=vn_fq_name)

    def create_port(self, vm_name, vmi_name, netns_name):
        vm_fq_name = ['_'.join(self.proj_fq_name + [vm_name])]
        if vm_fq_name not in self.vm_fq_names:
            vm = vnc_api.VirtualMachine(name='_'.join(vm_fq_name), fq_name=vm_fq_name)
            self.vnc_lib.virtual_machine_create(vm)
        vm = self.vnc_lib.virtual_machine_read(fq_name=vm_fq_name)

        vmi_fq_name = self.proj_fq_name + [vmi_name]
        if vmi_fq_name not in self.vmi_fq_names:
            vmi = vnc_api.VirtualMachineInterface(name=vmi_name, 
                                                  fq_name=vmi_fq_name, parent_obj=self.proj_obj)
            vmi.set_virtual_network(self.vn)
            vmi.set_virtual_machine(vm)
            self.vnc_lib.virtual_machine_interface_create(vmi)
        vmi = self.vnc_lib.virtual_machine_interface_read(fq_name=vmi_fq_name)

        inst_ip_fq_name = ['_'.join(self.proj_fq_name + [vmi_name])]
        if inst_ip_fq_name not in self.inst_ip_fq_names:
            ip = vnc_api.InstanceIp(name='_'.join(inst_ip_fq_name), fq_name=inst_ip_fq_name)
            ip.set_virtual_machine_interface(vmi)
            ip.set_virtual_network(self.vn)
            self.vnc_lib.instance_ip_create(ip)
        inst_ip = self.vnc_lib.instance_ip_read(fq_name=inst_ip_fq_name)

        ip = inst_ip.instance_ip_address
        mac = vmi.virtual_machine_interface_mac_addresses.mac_address[0]
        subnet = self.vn.network_ipam_refs[0]['attr'].ipam_subnets[0]
        gw = subnet.default_gateway
        dns = gw
        ip_prefix = '{}/{}'.format(ip, subnet.subnet.ip_prefix_len)

        if not self.args['config_only']:
            self.create_veth(self.vn, vm, vmi, netns_name, ip_prefix, gw, mac)

        object_dict = {
            'vn_name': self.vn.name,
            'vn_uuid': self.vn.uuid,
            'vm_name': vm.name,
            'vmi_name': vmi.name,
            'ip_prefix': ip_prefix,
            'mac': mac,
            'gw': gw,
            'dns': dns,
        }
        print(format_dict(object_dict, "table"))
    
    def create_veth(self, vn, vm, vmi, netns_name, ip_prefix, gw, mac):
        tap_name = mac.replace(':', '')[2:10] + '-' + mac.replace(':', '')[10:]
        tap_ns = "veth" + tap_name
        tap_host = "tap" + tap_name
        network = netaddr.IPNetwork(ip_prefix)
        ip = str(network.ip)
        ip_prefix = str(network)
        broadcast = str(network.broadcast)

        if netns_name not in self.netns_list:
            call(['ip', 'netns', 'add', netns_name])

        if tap_host not in self.tap_list:
            ns_exec = ['ip', 'netns', 'exec', netns_name]
            call(['ip', 'link', 'add', tap_host, 'type', 'veth', 'peer', 'name', tap_ns])
            call(['ip', 'link', 'set', tap_ns,'netns', netns_name])
            call(ns_exec + ['ip', 'link', 'set', 'dev', tap_ns, 'address', mac])
            call(ns_exec + ['ip', 'address', 'add', ip_prefix, 'broadcast', broadcast, 'dev', tap_ns])
            call(ns_exec + ['ip', 'link', 'set', 'dev', tap_ns, 'up'])
            call(ns_exec + ['route', 'add', 'default', 'gw', gw, 'dev', tap_ns])
            call(['ip', 'link', 'set', 'dev', tap_host, 'up'])

        port = instance_service.ttypes.Port(
            uuid_from_string(vmi.uuid),
            uuid_from_string(vm.uuid),
            tap_host,
            ip,
            uuid_from_string(vn.uuid),
            mac,
        )
        rpc = vrouter_rpc()
        rc = rpc.AddPort([port])

    def delete_all(self):
        vm_prefix = '_'.join(self.proj_fq_name) + '_'
        vm_fq_names = [x for x in self.vm_fq_names if x[0].startswith(vm_prefix)]
        for vm_fq_name in vm_fq_names:
            try:
                vm = self.vnc_lib.virtual_machine_read(fq_name=vm_fq_name)
                vmis = vm.get_virtual_machine_interface_back_refs()
                vmis = vmis if vmis else []
                for vmi in vmis:
                    try:
                        vmi = self.vnc_lib.virtual_machine_interface_read(id=vmi['uuid'])
                        try:
                            mac = vmi.virtual_machine_interface_mac_addresses.mac_address[0]
                            tap_name = 'tap' + mac.replace(':', '')[2:10] + '-' + mac.replace(':', '')[10:]
                            call(['ip', 'link', 'delete', tap_name])
                        except:
                            pass
                        try:
                            self.vrouter_api.delete_port(vmi.uuid)
                        except:
                            pass
                        inst_ips = vmi.get_instance_ip_back_refs()
                        inst_ips = inst_ips if inst_ips else []
                        for ip in inst_ips:
                            try:
                                print('Deleting {}'.format(ip['to']))
                                self.vnc_lib.instance_ip_delete(id=ip['uuid'])
                            except cfgm_common.exceptions.NoIdError:
                                pass
                        print('Deleting {}'.format(vmi.fq_name))
                        self.vnc_lib.virtual_machine_interface_delete(id=vmi.uuid)
                    except cfgm_common.exceptions.NoIdError:
                        pass
                self.vnc_lib.virtual_machine_delete(id=vm.uuid)
            except cfgm_common.exceptions.NoIdError:
                pass
        
        for vn_fq_name in self.vn_fq_names:
            if vn_fq_name[:2] == self.proj_fq_name:
                print('Deleting {}'.format(vn_fq_name))
                try:
                    self.vnc_lib.virtual_network_delete(fq_name=vn_fq_name)
                except cfgm_common.exceptions.NoIdError:
                    pass
                except cfgm_common.exceptions.RefsExistError as e:
                    print(str(e))

        for ipam_fq_name in self.ipam_fq_names:
            if ipam_fq_name[:2] == self.proj_fq_name: 
                print('Deleting {}'.format(ipam_fq_name))
                try:
                    self.vnc_lib.network_ipam_delete(fq_name=ipam_fq_name)
                except cfgm_common.exceptions.NoIdError:
                    pass
                except cfgm_common.exceptions.RefsExistError as e:
                    print(str(e))

        """
        print('Deleting {}'.format(self.proj_fq_name))
        try:
            self.vnc_lib.project_delete(fq_name=self.proj_fq_name)
        except cfgm_common.exceptions.NoIdError:
            pass
        """

        for netns_name in self.netns_list:
            call(['ip', 'netns', 'delete', netns_name])

        call(['service', 'supervisor-vrouter', 'restart'])

    def setup(self):
        if self.args['clear']:
            self.delete_all()
        else:
            project_name = self.args['project_name']
            vn_names = []
            super_net = netaddr.IPNetwork('10.0.0.0/8')
            for num_vn, subnet in enumerate(super_net.subnet(24, self.args['num_vn']), start=1):
                vn_name = 'vn_{:04d}'.format(num_vn)
                subnet_addr, subnet_len = str(subnet).split('/')
                self.create_vn(vn_name, subnet_addr, int(subnet_len))
                for num_port in range(1, self.args['num_port_per_vn'] + 1):
                    vm_name = 'vm_{}_{:04d}_{:02d}'.format(self.hostname, num_vn, num_port)
                    vmi_name = 'vmi_{}_{:04d}_{:02d}'.format(self.hostname, num_vn, num_port)
                    netns_name = '{}_{:04d}_{:02d}'.format(project_name, num_vn, num_port)
                    self.create_port(vm_name, vmi_name, netns_name)


if __name__ == '__main__':
    mock_port = MockPort()
    mock_port.setup()

